///////////////////////////////////////////////////////////////////////
// ezdib.C   - EZTWAIN library, DIB services module
// Copyright (C) 1999-2001 by Dosadi.  All Rights Reserved.
//
// Download from:    www.dosadi.com
//
// Support contact:  support@dosadi.com
//
// See end of eztwain.c for revision history
//
///////////////////////////////////////////////////////////////////////
//------------ Includes

#include <assert.h>
#include <memory.h>		// memset, memcpy
#include <stdio.h>      // sprintf
#include <io.h>         // for _lseek
#include <crtdbg.h>     // _CrtIsValidPointer
#include <time.h>
#include <math.h>       // fabs
#include "windows.h"

#ifdef _WIN32
#ifndef _INC_VFW
#include <vfw.h>
#endif
#else
#include "commdlg.h"
#endif

#include "twain.h"

#ifdef _WIN32
  #define EZTAPI WINAPI
#else
  #define EZTAPI __export FAR PASCAL
#endif

#include "eztwain.h"

///////////////////////////////////////////////////////////////////////
// Typedefs

typedef RGBQUAD FAR *LPQUAD;

///////////////////////////////////////////////////////////////////////
// Constants & Macros

#define STATIC static

#ifdef _WIN32
  #define INT32 int
  #define DATAGROUP unsigned
  #define DSM_FILENAME "TWAIN_32.DLL"
  #define DSM_ENTRYPOINT "DSM_Entry"
  #define IsValidHandle(h) (h!=NULL)
  #define HUGEWRITE(fh, pb, bc) ((INT32)_lwrite(fh, pb, bc))
  #define HUGEREAD(fh, pb, bc) ((INT32)_lread(fh, pb, bc))
  #define FMEMSET(p, v, n) memset(p, v, n);
  #define FMEMCPY(p, q, n) memcpy(p, q, n);
#else
  #define INT32 long
  #define DATAGROUP unsigned long
  #define DSM_FILENAME "TWAIN.DLL"
  #define DSM_ENTRYPOINT "DSM_ENTRY"
  #define VALID_HANDLE 32
  #define IsValidHandle(h) ((h!=0) && ((h) >= 32))
  #define HUGEWRITE(fh, pb, bc) _hwrite(fh, pb, bc)
  #define HUGEREAD(fh, pb, bc) _hread(fh, pb, bc)
  #define FMEMSET(p, v, n) _fmemset(p, v, n);
  #define FMEMCPY(p, q, n) _fmemcpy(p, q, n);
#endif

#define Lock(hdib) ((LPBITMAPINFOHEADER)GlobalLock(hdib))
#define Unlock(hdib) GlobalUnlock(hdib)


///////////////////////////////////////////////////////////////////////
// Imports from other modules

HANDLE EZTAPI TWAIN_LoadNativeFromFile(HFILE fh);
int WriteToJpeg(LPBITMAPINFOHEADER pbi, LPCSTR pszFile);
extern FILE* pfsLog;             // tracing log

///////////////////////////////////////////////////////////////////////
// Bitmap utilities

static double Integerize(double x)
// If x is *very* close to being an integer, round it.
// *very close* means within 0.01%
{
   long lx = (long)(x+0.5);
   double err = fabs(x - lx);
   if (x >= 50.0 && err < 0.01) {
       return lx;
   } else {
       return x;
   }
}


double PPM2DPI(long ppm)
	{ return 0.0254 * ppm; }

long DPI2PPM(double dpi)
	{ return ((long)(dpi / 0.0254 + 0.5)); }


static int ColorCount(int bpp)
{
    return (bpp > 8) ? 0 : (1 << bpp);
}

       
static size_t RowBytes(int bpp, int w)
{
	// bytes per image row - round up to DWORD
	return (bpp * w + 31) / 32 * 4;
}

static size_t ColorTableBytes(int bpp)
{
	return ColorCount(bpp) * sizeof(RGBQUAD);
} // ColorTableBytes



size_t ImageBytes(int bpp, int w, int h)
{
	return abs(h) * RowBytes(bpp, w);
} // ImageBytes


size_t DibBytes(int bpp, int w, int h)
{
	return sizeof(BITMAPINFOHEADER) +
		   ColorTableBytes(bpp) +
		   ImageBytes(bpp, w, h);
} // DibBytes



//---------------------------------------------------------------------------

const PALETTEENTRY peStock256[256] = {
   // from 0 .. 9 are Windows fixed colors
     0,   0,   0, 0,	// 000 always black
   128,   0,   0, 0,	// 001 dark red
     0, 128,   0, 0,	// 002 dark green
   128, 128,   0, 0,	// 003 dark brown
     0,   0, 128, 0,	// 004 dark blue
   128,   0, 128, 0,	// 005 dark purple
     0, 128, 128, 0,	// 006 dark teal
   192, 192, 192, 0,	// 007 light gray
   192, 220, 192, 0,	// 008 pale green
   166, 202, 240, 0,	// 009 sky blue
     4,   4,   4, 4,	// 010 dark gray ramp
     8,   8,   8, 4,	// 011
    12,  12,  12, 4,	// 012
    17,  17,  17, 4,	// 013
    22,  22,  22, 4,	// 014
    28,  28,  28, 4,	// 015
    34,  34,  34, 4,	// 016
    41,  41,  41, 4,	// 017
    85,  85,  85, 4,	// 018
    77,  77,  77, 4,	// 019
    66,  66,  66, 4,	// 020
    57,  57,  57, 4,	// 021
   255, 124, 128, 4,	// 022
   255,  80,  80, 4,	// 023
   214,   0, 147, 4,	// 024
   204, 236, 255, 4,	// 025
   239, 214, 198, 4,	// 026
   231, 231, 214, 4,	// 027
   173, 169, 144, 4,	// 028
    51,   0,   0, 4,	// 029
   102,   0,   0, 4,	// 030
   153,   0,   0, 4,	// 031 
   204,   0,   0, 4,	// 032
     0,  51,   0, 4,	// 033
	51,  51,   0, 4,	// 034
   102,  51,   0, 4,	// 035
   153,  51,   0, 4,	// 036
   204,  51,   0, 4,	// 037
   255,  51,   0, 4,	// 038
     0, 102,   0, 4,	// 039
    51, 102,   0, 4,	// 040
   102, 102,   0, 4,	// 041
   153, 102,   0, 4,	// 042
   204, 102,   0, 4,	// 043
   255, 102,   0, 4,	// 044
     0, 153,   0, 4,	// 045     
	51, 153,   0, 4,	// 046
   102, 153,   0, 4,	// 047
   153, 153,   0, 4,	// 048
   204, 153,   0, 4,	// 049
   255, 153,   0, 4,	// 050
     0, 204,   0, 4,	// 051
    51, 204,   0, 4,	// 052
   102, 204,   0, 4,	// 053
   153, 204,   0, 4,	// 054
   204, 204,   0, 4,	// 055
   255, 204,   0, 4,	// 056
   102, 255,   0, 4,	// 057
   153, 255,   0, 4,	// 058
   204, 255,   0, 4,	// 059
     0,   0,  51, 4,	// 060
    51,   0,  51, 4,	// 061
   102,   0,  51, 4,	// 062
   153,   0,  51, 4,	// 063
   204,   0,  51, 4,	// 064
   255,   0,  51, 4,	// 065
     0,  51,  51, 4,	// 066
    51,  51,  51, 4,	// 067
   102,  51,  51, 4,	// 068
   153,  51,  51, 4,	// 069
   204,  51,  51, 4,	// 070
   255,  51,  51, 4,	// 071
     0, 102,  51, 4,	// 072
    51, 102,  51, 4,	// 073
   102, 102,  51, 4,	// 074
   153, 102,  51, 4,	// 075
   204, 102,  51, 4,	// 076
   255, 102,  51, 4,	// 077
     0, 153,  51, 4,	// 078
    51, 153,  51, 4,	// 079
   102, 153,  51, 4,	// 080
   153, 153,  51, 4,	// 081
   204, 153,  51, 4,	// 082
   255, 153,  51, 4,	// 083
     0, 204,  51, 4,	// 084
    51, 204,  51, 4,	// 085
   102, 204,  51, 4,	// 086
   153, 204,  51, 4,	// 087
   204, 204,  51, 4,	// 088
   255, 204,  51, 4,	// 089  
    51, 255,  51, 4,	// 090
   102, 255,  51, 4,	// 091
   153, 255,  51, 4,	// 092
   204, 255,  51, 4,	// 093
   255, 255,  51, 4,	// 094
     0,   0, 102, 4,	// 095
    51,   0, 102, 4,	// 096
   102,   0, 102, 4,	// 097
   153,   0, 102, 4,	// 098
   204,   0, 102, 4,	// 099
   255,   0, 102, 4,	// 100
     0,  51, 102, 4,	// 101
    51,  51, 102, 4,	// 102
   102,  51, 102, 4,	// 103
   153,  51, 102, 4,	// 104
   204,  51, 102, 4,	// 105
   255,  51, 102, 4,	// 106
     0, 102, 102, 4,	// 107
    51, 102, 102, 4,	// 108
   102, 102, 102, 4,	// 109
   153, 102, 102, 4,	// 110
   204, 102, 102, 4,	// 111
     0, 153, 102, 4,	// 112
    51, 153, 102, 4,	// 113
   102, 153, 102, 4,	// 114
   153, 153, 102, 4,	// 115
   204, 153, 102, 4,	// 116
   255, 153, 102, 4,	// 117
     0, 204, 102, 4,	// 118
    51, 204, 102, 4,	// 119
   153, 204, 102, 4,	// 120
   204, 204, 102, 4,	// 121
   255, 204, 102, 4,	// 122
     0, 255, 102, 4,	// 123
    51, 255, 102, 4,	// 124
   153, 255, 102, 4,	// 125
   204, 255, 102, 4,	// 126
   255,   0, 204, 4,	// 127
   204,   0, 255, 4,	// 128
     0, 153, 153, 4,	// 129
   153,  51, 153, 4,	// 130
   153,   0, 153, 4,	// 131
   204,   0, 153, 4,	// 132
     0,   0, 153, 4,	// 133
    51,  51, 153, 4,	// 134
   102,   0, 153, 4,	// 135
   204,  51, 153, 4,	// 136
   255,   0, 153, 4,	// 137
     0, 102, 153, 4,	// 138
    51, 102, 153, 4,	// 139
   102,  51, 153, 4,	// 140
   153, 102, 153, 4,	// 141
   204, 102, 153, 4,	// 142
   255,  51, 153, 4,	// 143
    51, 153, 153, 4,	// 144
   102, 153, 153, 4,	// 145
   153, 153, 153, 4,	// 146
   204, 153, 153, 4,	// 147
   255, 153, 153, 4,	// 148
     0, 204, 153, 4,	// 149
    51, 204, 153, 4,	// 150
   102, 204, 102, 4,	// 151
   153, 204, 153, 4,	// 152
   204, 204, 153, 4,	// 153
   255, 204, 153, 4,	// 154
     0, 255, 153, 4,	// 155
    51, 255, 153, 4,	// 156
   102, 204, 153, 4,	// 157
   153, 255, 153, 4,	// 158
   204, 255, 153, 4,	// 159
   255, 255, 153, 4,	// 160
     0,   0, 204, 4,	// 161
    51,   0, 153, 4,	// 162
   102,   0, 204, 4,	// 163
   153,   0, 204, 4,	// 164
   204,   0, 204, 4,	// 165
     0,  51, 153, 4,	// 166
    51,  51, 204, 4,	// 167
   102,  51, 204, 4,	// 168
   153,  51, 204, 4,	// 169
   204,  51, 204, 4,	// 170
   255,  51, 204, 4,	// 171
     0, 102, 204, 4,	// 172
    51, 102, 204, 4,	// 173
   102, 102, 153, 4,	// 174
   153, 102, 204, 4,	// 175
   204, 102, 204, 4,	// 176
   255, 102, 153, 4,	// 177
     0, 153, 204, 4,	// 178
    51, 153, 204, 4,	// 179
   102, 153, 204, 4,	// 180
   153, 153, 204, 4,	// 181
   204, 153, 204, 4,	// 182
   255, 153, 204, 4,	// 183
     0, 204, 204, 4,	// 184
    51, 204, 204, 4,	// 185
   102, 204, 204, 4,	// 186
   153, 204, 204, 4,	// 187
   204, 204, 204, 4,	// 188
   255, 204, 204, 4,	// 189
     0, 255, 204, 4,	// 190
    51, 255, 204, 4,	// 191
   102, 255, 153, 4,	// 192
   153, 255, 204, 4,	// 193
   204, 255, 204, 4,	// 194
   255, 255, 204, 4,	// 195
    51,   0, 204, 4,	// 196
   102,   0, 255, 4,	// 197
   153,   0, 255, 4,	// 198
     0,  51, 204, 4,	// 199
    51,  51, 255, 4,	// 200
   102,  51, 255, 4,	// 201
   153,  51, 255, 4,	// 202
   204,  51, 255, 4,	// 203
   255,  51, 255, 4,	// 204 
     0, 102, 255, 4,	// 205
    51, 102, 255, 4,	// 206
   102, 102, 204, 4,	// 207
   153, 102, 255, 4,	// 208
   204, 102, 255, 4,	// 209
   255, 102, 204, 4,	// 210
     0, 153, 255, 4,	// 211
    51, 153, 255, 4,	// 212
   102, 153, 255, 4,	// 213
   153, 153, 255, 4,	// 214
   204, 153, 255, 4,	// 215
   255, 153, 255, 4,	// 216
     0, 204, 255, 4,	// 217
    51, 204, 255, 4,	// 218
   102, 204, 255, 4,	// 219
   153, 204, 255, 4,	// 220
   204, 204, 255, 4,	// 221
   255, 204, 255, 4,	// 222
    51, 255, 255, 4,	// 223
   102, 255, 204, 4,	// 224
   153, 255, 255, 4,	// 225
   204, 255, 255, 4,	// 226
   255, 102, 102, 4,	// 227
   102, 255, 102, 4,	// 228
   255, 255, 102, 4,	// 229
   102, 102, 255, 4,	// 230
   255, 102, 255, 4,	// 231
   102, 255, 255, 4,	// 232
   165,   0,  33, 4,	// 233
    95,  95,  95, 4,	// 234
   119, 119, 119, 4,	// 235 
   134, 134, 134, 4,	// 236
   150, 150, 150, 4,	// 237
   203, 203, 203, 4,	// 238
   178, 178, 178, 4,	// 239
   215, 215, 215, 4,	// 240
   221, 221, 221, 4,	// 241
   227, 227, 227, 4,	// 242
   234, 234, 234, 4,	// 243
   241, 241, 241, 4,	// 244
   248, 248, 248, 4,	// 245
   // 246 - 255 are Windows fixed colors    
   255, 251, 240, 0,	// 246
   160, 160, 164, 0,	// 247
   128, 128, 128, 0,	// 248
   255,   0,   0, 0,	// 249
     0, 255,   0, 0,	// 250
   255, 255,   0, 0,	// 251
     0,   0, 255, 0,	// 252
   255,   0, 255, 0,	// 253
     0, 255, 255, 0,	// 254
   255, 255, 255, 0,	// 255 always white 
};

///////////////////////////////////////////////////////////////////////
// Bmi* Functions - work on locked DIB data

int BmiHeight(const LPBITMAPINFOHEADER pbi)
{
   return abs(pbi->biHeight);
} // BmiHeight


static int BmiColorCount(LPBITMAPINFOHEADER pbi)
{
   if (pbi->biSize == sizeof(BITMAPCOREHEADER)) {
      LPBITMAPCOREHEADER    lpbc = ((LPBITMAPCOREHEADER)pbi);
      return 1 << lpbc->bcBitCount;
   } else if (pbi->biClrUsed == 0) {
      return ColorCount(pbi->biBitCount);
   } else {
      return (int)pbi->biClrUsed;
   }
} // BmiColorCount


static size_t BmiColorTableBytes(LPBITMAPINFOHEADER pbi)
{
	return BmiColorCount(pbi) * sizeof(RGBQUAD);
} // BmiColorTableBytes


static size_t OffsetToBits(LPBITMAPINFOHEADER pbi)
{
	return (size_t)(pbi->biSize + BmiColorTableBytes(pbi));
} // OffsetToBits


static LPBYTE FindBits(LPBITMAPINFOHEADER pbi)
{
	return (LPBYTE)pbi + OffsetToBits(pbi);
} // FindBits


LPBYTE BmiFindBits(LPBITMAPINFOHEADER pbi)
// Given a pointer to a locked DIB, return a pointer to the actual bits (pixels)
{
    DWORD dwColorTableSize = (DWORD)(BmiColorCount(pbi) * sizeof(RGBQUAD));
    return (LPBYTE)pbi + pbi->biSize + dwColorTableSize;
} // BmiFindBits


long BmiRowBytes(const LPBITMAPINFOHEADER pbi)
{
	return RowBytes(pbi->biBitCount, pbi->biWidth);
} // BmiRowBytes


static LPBYTE BmiFindRow(LPBITMAPINFOHEADER lpbi, int nRow)
{
	LPBYTE lpbits = FindBits(lpbi);

	if (lpbi->biHeight > 0) {
		// standard upside-down DIB
		nRow = (int)lpbi->biHeight - 1 - nRow;
	}
	return lpbits + nRow * BmiRowBytes(lpbi);
} // BmiFindRow


static RGBQUAD* BmiColorTable(const LPBITMAPINFOHEADER pbi)
// Locate and return a pointer to the color table of a bitmap.
{
	return (RGBQUAD*)((LPBYTE)pbi + pbi->biSize);
} // BmiColorTable


void BmiInitializeHeader(LPBITMAPINFOHEADER pbi, int w, int h, int bpp)
{
	pbi->biSize = sizeof (BITMAPINFOHEADER);
	pbi->biWidth = w;
	pbi->biHeight = h;
	pbi->biPlanes = 1;
	pbi->biBitCount = (WORD)bpp;
	pbi->biCompression = BI_RGB;
	pbi->biSizeImage = ImageBytes(bpp, w, h);
	pbi->biXPelsPerMeter = 0;
	pbi->biYPelsPerMeter = 0;
	pbi->biClrUsed = 0;
	pbi->biClrImportant = 0;
} // BmiInitializeHeader


void BmiSetGrayColorTable(LPBITMAPINFOHEADER pbi)
{
   RGBQUAD* pct = BmiColorTable(pbi);
   int nColors = BmiColorCount(pbi);
   int i;
	// Create a grayscale linear ramp color table
	for (i = 0; i < nColors; i++) {
		BYTE g = (BYTE)(i * 255 / (nColors-1));
		pct[i].rgbBlue =
			pct[i].rgbGreen =
				pct[i].rgbRed = g;
		pct[i].rgbReserved = 0;
	} // for
} // BmiSetGrayColorTable


static void BmiBugger(BYTE* p, int n)
// Swap 1 and 3rd bytes of n 3-byte pixels
// This converts RGB -> BGR or vice-versa.
{
   while (n--) {
      register BYTE b = p[0]; p[0] = p[2]; p[2] = b;
      p += 3;
   }
} // BmiBugger


static void BmiBuggerAll(LPBITMAPINFOHEADER pbi)
{
    int y;
    for (y = 0; y < BmiHeight(pbi); y++) {
        BmiBugger(BmiFindRow(pbi, y), pbi->biWidth);
    }
} // BmiBuggerAll


void EZTAPI BmiWrite(LPBITMAPINFOHEADER pbi, int x, int y, int w, int h, const BYTE* pdata, long nStep)
// Write a rectangle of pixels into DIB at pbi starting at pixel (x,y)
// with width = w and height = h.  Take the pixels from memory starting
// at pdata, stepping by nStep bytes to get to the next row.
// Pixels are assumed to occupy the same number of bytes in memory as in the DIB.
{
    // Offset in row to pixel position x:
    int dx = (x * pbi->biBitCount) / 8;
    // Byte count to copy:
    int bc = (w * pbi->biBitCount + 7) / 8;
    while (h > 0) {
        FMEMCPY(BmiFindRow(pbi, y) + dx, pdata, bc);
        y++; h--; pdata += nStep;
    }
} // BmiWrite


void EZTAPI BmiWriteChannel(LPBITMAPINFOHEADER pbi, int nChannel, int x, int y, int w, int h, const BYTE* pdata, long nStep)
// nChannel = 0 (Red), 1 (Green), 2 (Blue)
{
    int dx, bc, n;
    if (pbi->biBitCount < 24) {
        // Channels don't affect sub-RGB dibs
        BmiWrite(pbi, x, y, w, h, pdata, nStep);
        return;
    }
    // Offset in row to pixel position x:
    dx = (x * pbi->biBitCount) / 8 + (2 - nChannel);
    // Byte count to copy:
    bc = (w * pbi->biBitCount + 7) / 8;
    nStep -= w;
    while (h > 0) {
        LPBYTE pdst = BmiFindRow(pbi, y) + dx;
        for (n = w; n--; pdst += 3) {
            *pdst = *pdata++;
        }
        y++; h--; pdata += nStep;
    }
} // BmiWriteChannel


int BmiWriteToBmpFile(LPBITMAPINFOHEADER pbi, HFILE fh)
{
	BITMAPFILEHEADER		bfh;
	int						fOk = FALSE;
	int						nBPP = pbi->biBitCount;
	int						nColors = (int)pbi->biClrUsed;

	// figure out actual size of color table
	if (nColors == 0 && nBPP <= 8) {
		nColors = (1 << nBPP);
	}
	if (pbi->biCompression == BI_RGB) {
		// uncompressed bitmap, image size might not be set
		DWORD dwBytesPerRow = (((pbi->biWidth * nBPP) + 31) / 32) * 4;
		pbi->biSizeImage = dwBytesPerRow * pbi->biHeight;
	} else if (pbi->biSizeImage == 0) {
		// compressed bitmap, image size had damn well better be set!
		return FALSE;
		// This could be hacked around with something like this:
		//pbi->biSizeImage = GlobalSize((HANDLE)GlobalHandle(HIWORD(pbi)));
	}

	// Set up BMP header.
	bfh.bfType = 0x4d42;                // "BM"
	bfh.bfReserved1 = 0;
	bfh.bfReserved2 = 0;
	bfh.bfOffBits = sizeof(BITMAPFILEHEADER) +
					pbi->biSize +
					sizeof(RGBQUAD) * nColors;
	bfh.bfSize = bfh.bfOffBits + pbi->biSizeImage;

	if (_lwrite(fh, (LPCSTR)&bfh, sizeof bfh) == sizeof bfh) {

		INT32 towrite = bfh.bfSize - (INT32)sizeof bfh;

		if (HUGEWRITE(fh, (LPCSTR)pbi, towrite) == towrite) {
			fOk = TRUE;
		}
	}

	return fOk;
} // BmiWriteToBmpFile


///////////////////////////////////////////////////////////////////////
// Crucial helper function for memory transfers

int CopyMemoryBufferToDib(HANDLE hdib,
                          TW_IMAGEMEMXFER mx,
                          TW_IMAGEINFO info,
                          int nChannel)
{
   int y = mx.YOffset;
   int x = mx.XOffset;
   int nPlanes = (info.Planar ? info.SamplesPerPixel : 1);
   // compute bits per column
   int bpc = info.BitsPerPixel / nPlanes;
   // nRowStep is number of bytes between the start of each row
   // in the memory buffer.  (Not in the DIB)
   unsigned nRowStep = 0;
   BYTE* psrc;

#ifdef MEM_FUSSY
   if (mx.BytesWritten == 0 ||
       mx.BytesWritten != (mx.BytesPerRow * mx.Rows) ||
       (mx.BytesPerRow * 8) < (mx.Columns * bpc)) {
       // mx stride info is really screwed up.
       return FALSE;
   }
#endif
   if (mx.BytesWritten != 0 && (mx.BytesPerRow * mx.Rows) == mx.BytesWritten) {
       // mx.BytesPerRow looks pretty good - consistent.
       nRowStep = mx.BytesPerRow;
   } else {
       // mx.BytesWritten or mx.BytesPerRow looks fishy
       // compute bytes/row rounded up to DWORD multiple
       nRowStep = (mx.Columns * bpc + 31) / 32 * 4;
       if (nRowStep != mx.BytesPerRow) {
           nRowStep = (mx.Columns * bpc + 7) / 8;
           if (nRowStep != mx.BytesPerRow) {
               // mx.BytesPerRow makes no sense at all
               return FALSE;
           }
           // try our best to recover
           nRowStep = min(nRowStep, mx.BytesPerRow);
       }
   }
   // Make sure DS doesn't feed us rows or columns outside the image:
   mx.Rows = min(mx.Rows, (unsigned)(info.ImageLength - mx.YOffset));
   mx.Columns = min(mx.Columns, (unsigned)(info.ImageWidth - mx.XOffset));

   if (mx.Rows * mx.Columns == 0) {
       // Null operation
       return TRUE;
   }

   if (pfsLog) {
      fprintf(pfsLog, "using nRowStep = %u\n", nRowStep);
   }
   if (mx.Memory.Flags & TWMF_POINTER) {
       psrc = (BYTE*)mx.Memory.TheMem;
   } else {
      psrc = (BYTE*)GlobalLock((HANDLE)mx.Memory.TheMem);
      if (!psrc) {
          return FALSE;
      }
   }
   assert(psrc);
   if (!_CrtIsValidPointer(psrc, mx.Rows * nRowStep, FALSE)) {
      MessageBox(NULL,
                 "TWAIN device driver returned invalid memory pointer.",
                 "TWAIN Error",
                 MB_ICONEXCLAMATION | MB_OK);
   } else {
       // Copy memory buffer rows into DIB
       LPBITMAPINFOHEADER pbi = (LPBITMAPINFOHEADER)GlobalLock(hdib);
       if (info.Planar) {
           BmiWriteChannel(pbi, nChannel, x, y, mx.Columns, mx.Rows, psrc, nRowStep);
       } else {
           BmiWrite(pbi, x, y, mx.Columns, mx.Rows, psrc, nRowStep);
       }
       GlobalUnlock(hdib);
   }
   if (mx.Memory.Flags & TWMF_HANDLE) {
       GlobalUnlock((HANDLE)mx.Memory.TheMem);
   }
   return TRUE;
} // CopyMemoryBufferToDib


///////////////////////////////////////////////////////////////////////
// DIB_* Functions

int EZTAPI DIB_PutOnClipboard(HANDLE hdib)
{
    int nRes = 0;
    if (OpenClipboard(NULL)) {
        if (EmptyClipboard()) {
            SetClipboardData(CF_DIB, hdib);
            nRes = TRUE;
        }
        CloseClipboard();
    }
    return nRes;
} // DIB_PutOnClipboard


HANDLE EZTAPI DIB_GetFromClipboard(void)
{
    HANDLE hdib = NULL;
    if (OpenClipboard(NULL)) {
        hdib = DIB_Copy(GetClipboardData(CF_DIB));
        CloseClipboard();
    }
    return hdib;
} // DIB_GetFromClipboard


HANDLE EZTAPI DIB_Copy(HANDLE hdib)
{
    HANDLE hdibNew = NULL;
    if (hdib) {
        DWORD dwBytes = GlobalSize(hdib);
        hdibNew = GlobalAlloc(GMEM_MOVEABLE, dwBytes);
        if (hdibNew) {
            LPVOID pvOld = GlobalLock(hdib);
            if (pvOld) {
                LPVOID pvNew = (LPBYTE)GlobalLock(hdibNew);
                if (pvNew) {
                    FMEMCPY(pvNew, pvOld, dwBytes);
                    GlobalUnlock(hdibNew);
                } else {
                    GlobalFree(hdibNew); hdibNew = NULL;
                }
                GlobalUnlock(hdib);
            } else {
                GlobalFree(hdibNew); hdibNew = NULL;
            }
        }
    }
    return hdibNew;
} // DIB_Copy


void EZTAPI DIB_SetColorCount(HANDLE hdib, int n)
{
	LPBITMAPINFOHEADER pbi = Lock(hdib);
    if (n != BmiColorCount(pbi)) {
        pbi->biClrUsed = n;
    }
	Unlock(hdib);
} // DIB_SetColorCount


void EZTAPI DIB_SetGrayColorTable(HANDLE hdib)
{
	LPBITMAPINFOHEADER pbi = Lock(hdib);
	BmiSetGrayColorTable(pbi);
	Unlock(hdib);
} // DIB_SetGrayColorTable


void EZTAPI DIB_SetColorTableRGB(HANDLE hdib, int nEntry, int R, int G, int B)
{
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    int nColors = BmiColorCount(pbi);
    if (nEntry < nColors) {
        RGBQUAD* pq = BmiColorTable(pbi) + nEntry;
        pq->rgbBlue    = (BYTE)B;
        pq->rgbGreen   = (BYTE)G;
        pq->rgbRed     = (BYTE)R;
        pq->rgbReserved = 0;
    }
    Unlock(hdib);
} // DIB_SetColorTableRGB


int EZTAPI DIB_ColorTableR(HANDLE hdib, int i)
{
    int v = 0;
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    if (i < BmiColorCount(pbi)) {
        v = (BmiColorTable(pbi) + i)->rgbRed;
    }
    Unlock(hdib);
    return v;
}


int EZTAPI DIB_ColorTableG(HANDLE hdib, int i)
{
    int v = 0;
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    if (i < BmiColorCount(pbi)) {
        v = (BmiColorTable(pbi) + i)->rgbGreen;
    }
    Unlock(hdib);
    return v;
}


int EZTAPI DIB_ColorTableB(HANDLE hdib, int i)
{
    int v = 0;
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    if (i < BmiColorCount(pbi)) {
        v = (BmiColorTable(pbi) + i)->rgbBlue;
    }
    Unlock(hdib);
    return v;
}


void EZTAPI DIB_Free(HANDLE hdib)
{
    GlobalFree(hdib);
}


HANDLE EZTAPI DIB_Allocate(int nDepth, int nWidth, int nHeight)
// Allocate a DIB of the given depth and dimensions (in pixels).
// The contents of the DIB are undefined.
// Resolution is set to 0.0, use DIB_SetDPI to modify.
// DIBs are allocated only in standard depths: 1, 4, 8, or 24
// For depths <= 8, a grayscale 'ramp' palette is assumed. 
{
    HANDLE hdib;

    nHeight = abs(nHeight);

    if (nDepth == 16 || nDepth == 48) {
        // semi-standard weird depths: OK
    } else if (nDepth > 8) nDepth = 24;
    else if (nDepth > 4) nDepth = 8;
    else if (nDepth > 1) nDepth = 4;
    
    hdib = (HANDLE)GlobalAlloc(GMEM_MOVEABLE, DibBytes(nDepth, nWidth, nHeight));
    if (hdib) {
        LPBITMAPINFOHEADER pbi = Lock(hdib);
        BmiInitializeHeader(pbi, nWidth, nHeight, nDepth);
        BmiSetGrayColorTable(pbi);
        Unlock(hdib);
    }
    return hdib;
} // DIB_Allocate


int EZTAPI DIB_Depth(HANDLE hdib)
{
   int n = Lock(hdib)->biBitCount;
   Unlock(hdib);
   return n;
} // DIB_Depth


int EZTAPI DIB_ColorCount(HANDLE hdib)
{
	int nColors = BmiColorCount(Lock(hdib));
	Unlock(hdib);
    return nColors;
} // DIB_NumColors


int EZTAPI DIB_Width(HANDLE hdib)
{
	int n = abs(Lock(hdib)->biWidth);
	Unlock(hdib);
	return n;
} // DIB_Width


int EZTAPI DIB_Height(HANDLE hdib)
{
	int n = BmiHeight(Lock(hdib));
	Unlock(hdib);
	return n;
} // DIB_Height


size_t EZTAPI DIB_RowBytes(HANDLE hdib)
{
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    size_t s = RowBytes(pbi->biBitCount, pbi->biWidth);
    Unlock(hdib);
    return s;
} // DIB_RowBytes


BITMAPINFOHEADER* EZTAPI DIB_Lock(HANDLE hdib)
{
    return Lock(hdib);
} // DIB_Lock


void EZTAPI DIB_Unlock(HANDLE hdib)
{
    Unlock(hdib);
} // DIB_Unlock


void EZTAPI DIB_SetResolution(HANDLE hdib, double xdpi, double ydpi)
{
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    pbi->biXPelsPerMeter = DPI2PPM(xdpi);
    pbi->biYPelsPerMeter = DPI2PPM(ydpi);
    Unlock(hdib);
} // DIB_SetResolution


double EZTAPI DIB_YResolution(HANDLE hdib)
{
   LPBITMAPINFOHEADER pbi = Lock(hdib);
   double dpi = Integerize(PPM2DPI(pbi->biYPelsPerMeter));
   Unlock(hdib);
   return dpi;
} // DIB_YResolution


double EZTAPI DIB_XResolution(HANDLE hdib)
{
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    double dpi = Integerize(PPM2DPI(pbi->biXPelsPerMeter));
    Unlock(hdib);
    return dpi;
} // DIB_XResolution


void EZTAPI DIB_ReadRow(HANDLE hdib, int nRow, BYTE* prow)
{
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    FMEMCPY(prow, BmiFindRow(pbi, nRow), (size_t)BmiRowBytes(pbi));
    Unlock(hdib);
}


void EZTAPI DIB_ReadRowRGB(HANDLE hdib, int nRow, BYTE* prow)
{
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    LPBYTE psrc = BmiFindRow(pbi, nRow);
    RGBQUAD* pct = BmiColorTable(pbi);
    int w = pbi->biWidth;               // pixels per row
    BYTE b;
    int ci, nb = 0;
    switch (pbi->biBitCount) {
    case 1:
        while (w--) {
            if (nb == 0) {
                b = *psrc++; nb = 8;
            }
            ci = (b >> 7) & 1;
            b <<= 1;
            --nb;
            prow[0] = pct[ci].rgbRed;
            prow[1] = pct[ci].rgbGreen;
            prow[2] = pct[ci].rgbBlue;
            prow += 3;
        }
        break;
    case 4:
        while (w--) {
            if (nb == 0) {
                b = *psrc++; nb = 2;
            }
            ci = (b >> 4) & 0xF;
            b <<= 4;
            --nb;
            prow[0] = pct[ci].rgbRed;
            prow[1] = pct[ci].rgbGreen;
            prow[2] = pct[ci].rgbBlue;
            prow += 3;
        }
        break;
    case 8:
        while (w--) {
            ci = *psrc++;
            prow[0] = pct[ci].rgbRed;
            prow[1] = pct[ci].rgbGreen;
            prow[2] = pct[ci].rgbBlue;
            prow += 3;
        }
        break;
    case 24:
        while (w--) {
            prow[0] = psrc[2];
            prow[1] = psrc[1];
            prow[2] = psrc[0];
            prow += 3; psrc += 3;
        } // while
        break;
    case 16:
    case 32:
    default:
        break;
    } // switch
    Unlock(hdib);
} // DIB_ReadRowRGB


void EZTAPI DIB_WriteRow(HANDLE hdib, int r, const BYTE* pdata)
{
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    FMEMCPY(BmiFindRow(pbi, r), pdata, (size_t)BmiRowBytes(pbi));
    Unlock(hdib);
} // DIB_WriteRow


void EZTAPI DIB_WriteRowChannel(HANDLE hdib, int r, const BYTE* psrc, int nChannel)
// nChannel = 0 (Red), 1 (Green), 2 (Blue)
{
    LPBITMAPINFOHEADER pbi = Lock(hdib);
    if (pbi->biBitCount == 24) {
        int nBytes = pbi->biWidth;
        LPBYTE pdst = BmiFindRow(pbi, r) + (2-nChannel);
        while (nBytes--) {
            *pdst = *psrc++;
            pdst += 3;
        }
    } else {
        // Channels don't affect sub-RGB dibs
        int nBytes = BmiRowBytes(pbi);
        FMEMCPY(BmiFindRow(pbi, r), psrc, nBytes);
    }
    Unlock(hdib);
} // DIB_WriteRowChannel


void EZTAPI DIB_SwapRedBlue(HANDLE hdib)
{
	LPBITMAPINFOHEADER	pbi = Lock(hdib);
   if (pbi) {
      if (pbi->biBitCount == 24) {
         BmiBuggerAll(pbi);
      }
      Unlock(hdib);
   }
} // DIB_SwapRedBlue


HPALETTE EZTAPI DIB_CreatePalette(HANDLE hdib)
{
    HPALETTE            hPalette = NULL;

	LPBITMAPINFOHEADER	pbi = Lock(hdib);

    if (pbi) {
	    int nColors = BmiColorCount(pbi);      		// size of DIB palette
		int nEntries = nColors ? nColors : 256;  	    // size of palette to create
		// allocate logical palette structure
	    LOGPALETTE *pPal = (LOGPALETTE*)LocalAlloc(LPTR,
					sizeof(LOGPALETTE) + nEntries * sizeof(PALETTEENTRY));
		if (pPal) {
			// Fill in the palette entries
			pPal->palNumEntries = (WORD)nEntries;
			pPal->palVersion    = 0x300;			// Windows 3.0 or later
			if (nColors) {
				// from the DIB color table
				// Get a pointer to the color table
				LPQUAD pRgb = (LPQUAD)((LPBYTE)pbi + (WORD)pbi->biSize);
			    WORD i;	
				// copy from DIB palette (triples, by the way) into the LOGPALETTE
				for (i = 0; i < nEntries; i++) {
					pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
					pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
					pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
					pPal->palPalEntry[i].peFlags = (BYTE)0;
				} // for
			} else {
				// Deep Dib: Synthesize halftone palette
				memcpy(pPal->palPalEntry, peStock256, sizeof peStock256);
			}
	
			// create a logical palette
			hPalette = CreatePalette(pPal);
			LocalFree((HANDLE)pPal);
		}
	
		Unlock(hdib);
	}
    return hPalette;
} // DIB_CreatePalette


void EZTAPI DIB_DrawToDC(HANDLE hdib, HDC hDC,
                         int dx, int dy,
                         int w,  int h,
                         int sx, int sy)
{
	LPBITMAPINFOHEADER pbi = Lock(hdib);
	if (pbi) {
#ifdef _WIN32
		HDRAWDIB hdd = DrawDibOpen();
		if (hdd) {
			DrawDibDraw(hdd, hDC,
						dx, dy, w, h,
						pbi, BmiFindBits(pbi),
						sx, sy,
                        pbi->biWidth, pbi->biHeight,
						0	
						);
			DrawDibClose(hdd);
		}
#else
		StretchDIBits(hDC,
					dx, dy,			// x,y destination
					w, h,			// pixel width & height in destination
					sx, sy,			// x,y source
					w, h,			// width & height in source
					BmiFindBits(pbi), // pointer to 'bits' - the pixels
					(LPBITMAPINFO)pbi,
					DIB_RGB_COLORS,	// DIB palette is (presumably?) RGB entries
					SRCCOPY);		// raster operation (copy)
#endif
		Unlock(hdib);
	}
} // DIB_DrawToDC


int EZTAPI DIB_WriteToBmp(HANDLE hdib, LPCSTR pszFile)
{
	int result = -2;  // file open error (invalid path or name, or access denied)
    OFSTRUCT ofs;
    HFILE fh = OpenFile(pszFile , &ofs, OF_CREATE | OF_WRITE | OF_SHARE_EXCLUSIVE);
    if (fh != HFILE_ERROR) {
        result = DIB_WriteToBmpFile(hdib, fh);
        _lclose(fh);
    }
    return result;
} // DIB_WriteToBmp


int EZTAPI DIB_WriteToBmpFile(HANDLE hdib, HFILE fh)
{
   int result = -3;
	LPBITMAPINFOHEADER pbi = Lock(hdib);
	if (pbi) {
		result = -4;
		if (BmiWriteToBmpFile(pbi, fh)) {
			result = 0;			// success
		}
		Unlock(hdib);
	}
	return result;
} //  TWAIN_WriteNativeToFile


HANDLE EZTAPI DIB_LoadFromFilename(LPCSTR pszFile)
{
	HANDLE hdib = NULL;
	char szFile[_MAX_PATH];
	HFILE fh;
	OFSTRUCT ofs;

	if (!pszFile || !*pszFile) {
		// prompt for filename
		OPENFILENAME ofn;
		FMEMSET(&ofn, 0, sizeof ofn);
		szFile[0] = '\0';
		ofn.lStructSize = sizeof(OPENFILENAME);
		ofn.hwndOwner = NULL;
        if (TWAIN_IsJpegAvailable()) {
            ofn.lpstrFilter = "Image Files (*.bmp,*.jpg)\0*.bmp;*.jpg\0Windows Bitmaps (*.bmp)\0*.bmp\0JPEG/JFIF (*.jpg)\0*.jpg\0\0";
            //ofn.lpstrDefExt = "bmp";
        } else {
            ofn.lpstrFilter = "Windows Bitmaps (*.bmp)\0*.bmp\0\0";
            ofn.lpstrDefExt = "bmp";
        }
		ofn.lpstrFile= szFile;
		ofn.nMaxFile = sizeof(szFile) - 5;
		ofn.Flags = OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY |
					OFN_NOREADONLYRETURN;
	
		if (!GetOpenFileName(&ofn)) {
			return NULL;					// user cancelled dialog
		}
		pszFile = szFile;
	}

	fh = OpenFile(pszFile, &ofs, OF_READ | OF_SHARE_DENY_WRITE);
	if (fh != HFILE_ERROR) {
		hdib = TWAIN_LoadNativeFromFile(fh);
		_lclose(fh);
	}
	return hdib;
} // DIB_LoadFromFilename


int EZTAPI DIB_WriteToJpeg(HANDLE hdib, LPCSTR pszFile)
{
   int nResult = -3;       // (unable to lock DIB handle)
   LPBITMAPINFOHEADER pbi = Lock(hdib);
   if (pbi) {
      nResult = WriteToJpeg(pbi, pszFile);
      Unlock(hdib);
   }
   return nResult;
} // DIB_WriteToJpeg


///////////////////////////////////////////////////////////////////////
// Obsoleted DIB helpers

int EZTAPI TWAIN_DibDepth(HANDLE hdib)
{
   int D = Lock(hdib)->biBitCount;
   Unlock(hdib);
   return D;
} // TWAIN_DibDepth

int EZTAPI TWAIN_DibWidth(HANDLE hdib)
{
   int W = (int)(Lock(hdib)->biWidth);
   Unlock(hdib);
   return W;
} // TWAIN_DibWidth


int EZTAPI TWAIN_DibHeight(HANDLE hdib)
{
   int H = BmiHeight(Lock(hdib));
   Unlock(hdib);
   return H;
} // TWAIN_DibHeight


int EZTAPI TWAIN_DibNumColors(HANDLE hdib)
// given a DIB handle, return the number of palette entries: 0,2,16, or 256
{
    return DIB_ColorCount(hdib);
}


long EZTAPI TWAIN_DibRowBytes(HANDLE hdib)
{
    return DIB_RowBytes(hdib);
}


void EZTAPI TWAIN_DibReadRow(HANDLE hdib, int nRow, BYTE* prow)
{
    DIB_ReadRow(hdib, nRow, prow);
}


HPALETTE EZTAPI TWAIN_CreateDibPalette (HANDLE hdib)
{
    return DIB_CreatePalette(hdib);
} // TWAIN_CreateDibPalette


void EZTAPI TWAIN_DrawDibToDC(HDC hDC, int dx, int dy, int w, int h,
							  HANDLE hdib, int sx, int sy)
{
    DIB_DrawToDC(hdib, hDC, dx, dy, w, h, sx, sy);
} // TWAIN_DrawDibToDC

